module Sem06 where


-- Разминка. Пусть есть тип бинарного дерева

data Tree a = Leaf a | Branch (Tree a) a (Tree a) --deriving Show

-- 1. Напишите для него представителя Eq и Functor

instance Eq a => Eq (Tree a) where
    (==) (Leaf a) (Leaf b) = a == b
    (==) (Branch l v r) (Branch l' v' r') = v == v' && l == l' && r == r'
    (==) _ _ = False

-- fmap id = id
-- fmap (f . g) = fmap f . fmap g

instance Functor Tree where
    fmap f (Leaf a) = Leaf (f a)
    fmap f (Branch l v r) = Branch (fmap f l) (f v) (fmap f r)

-- 2. Напишите для него функцию elemTree, которая бы проверяла, принадлежит ли данный элемент дереву

elemTree :: Eq a => a -> Tree a -> Bool
elemTree x (Leaf y) = x == y
elemTree x (Branch l v r) = v == x || elemTree x l || elemTree x r


-- 3. Работает ли эта функция на бесконечных деревьях? Что нужно сделать, чтобы она работала?

infiniteTree :: Tree Integer
infiniteTree = Branch infiniteTree 1 infiniteTree

-- Класс типов Show. Нужен, чтобы переводить значения ваших типов данных в строковое представление.
-- Объявление(стащенное из хугла):

-- class  Show a  where
--     showsPrec :: Int    -- ^ the operator precedence
--               -> a      -- ^ the value to be convertd to a 'String'
--               -> ShowS
--
--     -- | A specialised variant of 'showsPrec', using precedence context
--     -- zero, and returning an ordinary 'String'.
--     show      :: a   -> String

-- Минимальное полное определение, как видно, либо show, либо showsPrec

-- showsPrec _ x s = show x ++ s
-- show x = shows x ""

-- shows :: (Show a) => ShowS
-- shows = showsPrec 0

-- Выше, когда я объявлял тип данных для дерева, я потребовал что бы у него
-- был реализован представитель Show по умолчанию. Представитель по умолчанию просто переведет
-- все конструкторы в строковое представление. Можно объявить своего представителя Show.
-- Сделаем это, например, для списков. Вариант первый:

data List a = Nil | Cons a (List a)
--
-- instance Show a => Show (List a) where
--     show Nil = ";"
--     show (Cons x Nil) = show x ++ ";"
--     show (Cons x xs) = show x ++ "," ++ show xs

-- Он не то, что бы эффективен из-за конкатенаций.
-- Давайте воспользуемся аккумулирующим параметром и напишем что-то более элегантное

myShowsList :: Show a => List a -> String -> String
myShowsList Nil acc = ';' : acc
myShowsList (Cons x Nil) acc = shows x (';' : acc)
myShowsList (Cons x xs) acc = shows x (',' : myShowsList xs acc)

-- Избавились от конкатенаций, теперь давайте введем(это уже сделано за нас на самом деле)
-- синоним для типа String -> String и избавимся от аккумулирующего параметра в явном виде
-- NB: я пишу штрихи, чтобы мой код не конфликтова с тем, что уже есть в Prelude

type ShowS' = String -> String

myShowsList' :: Show a => List a -> ShowS'
myShowsList' Nil = showChar ';'
myShowsList' (Cons x Nil) = shows x . showChar ';'
myShowsList' (Cons x xs) = shows x . showChar ',' . myShowsList' xs

-- Мы вроде ничего особо такого не сделали, но надо отметить, что мы перешли от комбинирования
-- объектов к комбинированию функций, что на самом деле многого стоит

-- напишите версию представителя Show для списков так, что бы списки выводились в виде
-- GHCi> (Cons 1 (Cons 2 (Cons 3 Nil)))
-- GHCi> "<1<2<3|>>>"

myShowsList'' :: Show a => List a -> ShowS
myShowsList'' Nil = showChar '|'
myShowsList'' (Cons x xs) = showChar '<' . shows x . myShowsList'' xs . showChar '>'

instance Show a => Show (List a) where
    showsPrec _ = myShowsList''

-- Сделайте то же самое, но для дерева
-- Например:
-- GHCi> (Branch (Leaf 1) 2 (Leaf 3))
-- GHCi> "<1{2}3>"

myShowsTree :: Show a => Tree a -> ShowS
myShowsTree (Leaf a) = shows a
myShowsTree (Branch l v r) = showChar '<' . myShowsTree l . showChar '{' . shows v . showChar '}' . myShowsTree r . showChar '>'

instance Show a => Show (Tree a) where
    showsPrec _ = myShowsTree

-- Про первый параметр в showsPrec

-- Как мы видели из объявления класса типов Show, первый параметр у showsPrec -- какое-то число
-- Оно нужно, что бы управлять выводом скобочек в зависимости от приоритета операторов
-- Например

infixr 7 :*:
infixr 6 :+:

data Cartesian6 a b = a :+: b --deriving Show
data Cartesian7 a b = a :*: b --deriving Show

-- механизм deriving генерирует примерно вот такую реализацию для этих двух типов:

instance (Show a, Show b) => Show (Cartesian7 a b) where
    showsPrec p (a :*: b) = showParen (p > astPrec) innerShows
        where
        astPrec = 7
        innerShows = showsPrec (astPrec + 1) a . showString " :*: " . showsPrec (astPrec + 1) b

instance (Show a, Show b) => Show (Cartesian6 a b) where
    showsPrec p (a :+: b) = showParen (p > plusPrec) innerShows
        where
        plusPrec = 6
        innerShows = showsPrec (plusPrec + 1) a . showString " :+: " . showsPrec (plusPrec + 1) b

-- Класс типов Read. Потребен для решения обратной задачи -- для преобразования строк в значения типа
-- Объявление с хугла:

-- class Read a where
--   readsPrec    :: Int   -- ^ the operator precedence of the enclosing
--                         -- context (a number from @0@ to @11@).
--                         -- Function application has precedence @10@.
--                 -> ReadS a
--   readPrec     :: ReadPrec a
--
--reads :: Read a => ReadS a

-- Минимальное полное определение readsPrec или readPrec. Второй вроде посвежее и более эффективен.
-- Мы посмотрим на первый

-- Как и в случае с классом Show, тут есть какой-то стремный тип ReadS a. Это просто синоним для стрелки:

type ReadS' a = String -> [(a, String)]

-- О нем можно думать, как о парсере, результат работы которого -- это значение типа a, и строка,
-- следующая за тем, что удалось разобрать. Неудачный разбор -- пустой список.
-- Обычно оно возвращает одноэлементный список

-- Простейший пример использования:
-- GHCi> (reads "99 bottles of beer") :: [(Integer, String)]
-- GHCi> [(99, " bottles of beer")]

-- можно написать представителя Read для нашего типа списков

myReadsList :: Read a => ReadS (List a)
myReadsList ('|' : s) = [(Nil, s)]
myReadsList ('<' : s) = [(Cons x xs, r) | (x, t) <- reads s, (xs, '>':r) <- myReadsList t]

instance Read a => Read (List a) where
    readsPrec _ = myReadsList


-- Сделайте то же самое, но для деревьев
-- Например:
-- GHCi> (myReadTree "<1{2}3> blah-blah") :: [(Tree Int, String)]
-- GHCi> [(Branch (Leaf 1) 2 (Leaf 3), " blah-blah")]

-- <<1{3}5>{2}3>

myReadTree :: Read a => ReadS (Tree a)
myReadTree ('<' : s) = [(Branch l v r, u) | (l, '{':t) <- myReadTree s, (v, '}' : t1) <- reads t, (r, '>' : u) <- myReadTree t1]
myReadTree cs = [(Leaf v, u) | (v, u) <- reads cs]


-- Ссылка на мягкое введение в хаскел -- http://rsdn.org/article/haskell/haskell_part2.xml#EOD
